# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
import numpy as np


def tensor_to_np(tensor: torch.Tensor) -> np.ndarray:
    return tensor.cpu().detach().numpy()


def get_mask(pad_size: torch.Size, lengths: torch.Tensor) -> torch.Tensor:
    device = lengths.device
    return torch.arange(pad_size, device=device)[None, :] < lengths[
        :, None
    ].to(device)
